#include <cstdlib>
#include <stdexcept>
#include <vector>
#include <iostream>
#include "lock_file.h"
#include "xunix.h"
#include "xutil.h"

namespace {

  const char* progname;
  static const char* progver = "1.0";
  bool debug = false;
  std::string debug_tag;

  void usage(std::ostream& ostrm)
  {
    ostrm << "\n";
    ostrm << "Usage: " << progname << " [<options>] <filename>" << "\n";
    ostrm << "\n";
    ostrm << "        Options:\n";
    ostrm << "\n";
    ostrm << "         -d <tag> or --debug <tag> : Debug.\n";
    ostrm << "         -h       or --help        : Help.\n";
    ostrm << "         -v       or --version     : Version.\n";
    ostrm << "\n";
    ostrm << "    \"" << progname << "\" is designed to work within the "
          << "context of a shell\n";
    ostrm << "    script.  You simply start it and wait via \"read\" for\n";
    ostrm << "    it to print its process id to standard out.  This\n";
    ostrm << "    indicates that it has locked the specified file for you\n";
    ostrm << "    using an advisory file lock that works on local files\n";
    ostrm << "    and NFS files (if the NFS server supports file locking).\n";
    ostrm << "    It will hold the lock until it is killed.\n";
    ostrm << "\n";
    ostrm << "Example:\n";
    ostrm << "\n";
    ostrm << "      slock <lock_file_name> | { read slock_pid\n";
    ostrm << "          # Critical section goes here.\n";
    ostrm << "          kill $slock_pid\n";
    ostrm << "      }\n";
    ostrm << std::endl;
  }

  std::string parse_command_line(int argc, char** argv)
  {
    std::vector<char*> non_flags;

    for( int i = 1 ; i < argc ; ++i ) {

      std::string arg(argv[i]);

      //
      // Obey the convention that "--" means to stop processing
      // command-line flags.
      //
      if( arg == "--" ) {
        // Copy the remaining parameters to the "non_flags" vector.
        for( int j = i + 1 ; j < argc ; ++j ) {
          non_flags.push_back(argv[j]);
        }
        break;
      }

      //
      // Search for valid flags.
      //
      if( (arg == "-d") || (arg == "--debug") ) {
        debug = true;
        ++i;
        if( i < argc ) {
          debug_tag = argv[i];
        } else {
          std::cerr << "\n";
          std::cerr << progname << ": Error: -d flag requires a tag.\n";
          usage(std::cerr);
          exit(1);
        }
        continue;
      }
      if( (arg == "-h") || (arg == "--help") ) {
        usage(std::cout);
        exit(0);
      }
      if( (arg == "-v") || (arg == "--version") ) {
        std::cout << progname << ": v" << progver << std::endl;
        exit(0);
      }

      // Search for invalid flags.
      if( (arg.length() > 0) && (arg[0] == '-') ) {
        std::cerr << "\n";
        std::cerr << progname << "Error: "
                  << "Invalid parameter at \"" << argv[i] << "\"\n";
        usage(std::cerr);
        exit(1);
      }

      // If it has survived the gamut, treat it as a non_flag.
      non_flags.push_back(argv[i]);
    }

    if( non_flags.size() == 0 ) {
      std::cerr << "\n";
      std::cerr << progname << ": Error: Lock file not specified\n";
      usage(std::cerr);
      exit(1);
    }

    if( non_flags.size() > 1 ) {
      std::cerr << "\n";
      std::cerr << progname << ": Error: Only one lock file allowed.\n";
      usage(std::cerr);
      exit(1);
    }

    return non_flags[0];
  }

  void do_nothing(int sig)
  {
    // empty
  }

}

int main(int argc, char* argv[])
{
  int rv = 0;

  try {

    progname = xutil::basename(argv[0]);

    // Receiving the following signals will cause us to gracefully
    // unlock the file.
    std::vector<int> exit_signals_vector;
    exit_signals_vector.push_back(SIGHUP);
    exit_signals_vector.push_back(SIGINT);
    exit_signals_vector.push_back(SIGQUIT);
    exit_signals_vector.push_back(SIGTERM);

    // The user passes to us on the command line the file name of the
    // file to lock.
    std::string filename = parse_command_line(argc, argv);
    
    // Lock the file.
    lock_file lf(filename);
    if( debug ) {
      std::cerr << progname << " (" << getpid() << "): "
                << "Waiting for \"" << filename << "\" "
                << "for \"" << debug_tag << "\"."
                << std::endl;
    }

    lf.lock();

    //
    // Until this point, allow an incoming signals default behavior.
    // After this point but before writing our pid to stdout you need
    // to be more careful with your signal handling in order to avoid
    // a race between when the user receives our pid and sends us an
    // exit signal and when we are waiting to receive the signal.
    //

    // Initialize the block_signals_set.
    sigset_t block_signals_set;
    xunix::sigemptyset(&block_signals_set);
    for( unsigned i = 0 ; i < exit_signals_vector.size() ; ++i ) {
      xunix::sigaddset(&block_signals_set, exit_signals_vector[i]);
    }

    // Initialize the wait_signals_set.  Remember, the signals that
    // are in the mask will be blocked during the call to
    // sigsuspend().
    sigset_t wait_signals_set;
    xunix::sigfillset(&wait_signals_set);
    for( unsigned i = 0 ; i < exit_signals_vector.size() ; ++i ) {
      xunix::sigdelset(&wait_signals_set, exit_signals_vector[i]);
    }

    // Avoid the usual race condition when waiting for one or more
    // signals.
    xunix::sigprocmask(SIG_BLOCK, &block_signals_set, 0);

    for( unsigned i = 0 ; i < exit_signals_vector.size() ; ++i ) {
      xutil::install_signal_handler(exit_signals_vector[i], & do_nothing);
    }

    // Don't print out the debug message until you know its
    // bookend-message of "Lock released" is (more-or-less)
    // guaranteed to also be printed.
    if( debug ) {
      std::cerr << progname << " (" << getpid() << "): "
                << "Lock grabbed "
                << "for \"" << debug_tag << "\"."
                << std::endl;
    }

    //
    // Now you're ready to tell the user your pid so that the user may
    // send an exit signal whenever she wishes.
    //

    // The process id of this program is the only thing that is ever
    // written to stdout during normal operation.  You can use this
    // start you shell script running by piping the output into
    // "read", and because you have read the process id of this
    // program, you shouldn't have a hard time killing it.
    std::cout << getpid() << std::endl;


    // Wait for one of the blocked signals to be received.
    xunix::sigsuspend(&wait_signals_set);


    if( debug ) {
      std::cerr << progname << " (" << getpid() << "): "
                << "Lock released "
                << "for \"" << debug_tag << "\"."
                << std::endl;
    }

    // Unlock the file.
    lf.unlock();

  } catch( const std::exception& e ) {
    std::cerr << progname << ": " << e.what() << std::endl;
    rv = 1;
  } catch( ... ) {
    std::cerr << progname
              << ": Error: Caught unexpected exception."
              << std::endl;
    rv = 1;
  }

  return rv;
}
